@@ -5,6 +5,7 @@
    libraries.
 
    Copyright (C) 2014, The University of Texas at Austin
+   Copyright (C) 2019, Advanced Micro Devices, Inc.
 
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
@@ -48,6 +49,10 @@ void bli_obj_create
 
 	bli_obj_create_without_buffer( dt, m, n, obj );
 
+	#ifdef BLIS_ENABLE_MEM_TRACING
+	printf( "bli_obj_create(): " );
+	#endif
+
 	bli_obj_alloc_buffer( rs, cs, 1, obj );
 }
 
@@ -232,6 +237,10 @@ void bli_obj_create_1x1
 {
 	bli_obj_create_without_buffer( dt, 1, 1, obj );
 
+	#ifdef BLIS_ENABLE_MEM_TRACING
+	printf( "bli_obj_create_1x1(): " );
+	#endif
+
 	bli_obj_alloc_buffer( 1, 1, 1, obj );
 }
 
@@ -277,7 +286,13 @@ void bli_obj_free
 		// is a detached scalar (ie: if the buffer pointer refers to the
 		// address of the internal scalar buffer).
 		if ( bli_obj_buffer( obj ) != bli_obj_internal_scalar_buffer( obj ) )
+		{
+			#ifdef BLIS_ENABLE_MEM_TRACING
+			printf( "bli_obj_free(): " );
+			#endif
+
 			bli_free_user( bli_obj_buffer( obj ) );
+		}
 	}
 }
 
@@ -390,7 +405,8 @@ void bli_adjust_strides
 	// matrix).
 	if ( m == 0 || n == 0 ) return;
 
-	// Interpret rs = cs = 0 as request for column storage.
+	// Interpret rs = cs = 0 as request for column storage and -1 as a request
+	// for row storage.
 	if ( *rs == 0 && *cs == 0 && ( *is == 0 || *is == 1 ) )
 	{
 		// First we handle the 1x1 scalar case explicitly.
@@ -399,8 +415,9 @@ void bli_adjust_strides
 			*rs = 1;
 			*cs = 1;
 		}
-		// We use column-major storage, except when m == 1, because we don't
-		// want both strides to be unit.
+		// We use column-major storage, except when m == 1, in which case we
+		// use what amounts to row-major storage because we don't want both
+		// strides to be unit.
 		else if ( m == 1 && n > 1 )
 		{
 			*rs = n;
@@ -430,6 +447,46 @@ void bli_adjust_strides
 		                                 BLIS_HEAP_STRIDE_ALIGN_SIZE );
 		}
 	}
+	else if ( *rs == -1 && *cs == -1 && ( *is == 0 || *is == 1 ) )
+	{
+		// First we handle the 1x1 scalar case explicitly.
+		if ( m == 1 && n == 1 )
+		{
+			*rs = 1;
+			*cs = 1;
+		}
+		// We use row-major storage, except when n == 1, in which case we
+		// use what amounts to column-major storage because we don't want both
+		// strides to be unit.
+		else if ( n == 1 && m > 1 )
+		{
+			*rs = 1;
+			*cs = m;
+		}
+		else
+		{
+			*rs = n;
+			*cs = 1;
+		}
+
+		// Use default complex storage.
+		*is = 1;
+
+		// Align the strides depending on the tilt of the matrix. Note that
+		// scalars are neither row nor column tilted. Also note that alignment
+		// is only done for rs = cs = -1, and any user-supplied row and column
+		// strides are preserved.
+		if ( bli_is_col_tilted( m, n, *rs, *cs ) )
+		{
+			*cs = bli_align_dim_to_size( *cs, elem_size,
+			                             BLIS_HEAP_STRIDE_ALIGN_SIZE );
+		}
+		else if ( bli_is_row_tilted( m, n, *rs, *cs ) )
+		{
+			*rs = bli_align_dim_to_size( *rs, elem_size,
+		                                 BLIS_HEAP_STRIDE_ALIGN_SIZE );
+		}
+	}
 	else if ( *rs == 1 && *cs == 1 )
 	{
 		// If both strides are unit, this is probably a "lazy" request for a
