@@ -6,7 +6,7 @@
 
    Copyright (C) 2014, The University of Texas at Austin
    Copyright (C) 2016, Hewlett Packard Enterprise Development LP
-   Copyright (C) 2018, Advanced Micro Devices, Inc.
+   Copyright (C) 2019, Advanced Micro Devices, Inc.
 
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
@@ -197,6 +197,16 @@ typedef double    f77_double;
 typedef scomplex  f77_scomplex;
 typedef dcomplex  f77_dcomplex;
 
+// -- Void function pointer types --
+
+// Note: This type should be used in any situation where the address of a
+// *function* will be conveyed or stored prior to it being typecast back
+// to the correct function type. It does not need to be used when conveying
+// or storing the address of *data* (such as an array of float or double).
+
+//typedef void (*void_fp)( void );
+typedef void* void_fp;
+
 
 //
 // -- BLIS info bit field offsets ----------------------------------------------
@@ -594,10 +604,10 @@ typedef enum
 	BLIS_SUBPART0,
 	BLIS_SUBPART1,
 	BLIS_SUBPART2,
-	BLIS_SUBPART1T,
+	BLIS_SUBPART1AND0,
+	BLIS_SUBPART1AND2,
+	BLIS_SUBPART1A,
 	BLIS_SUBPART1B,
-	BLIS_SUBPART1L,
-	BLIS_SUBPART1R,
 	BLIS_SUBPART00,
 	BLIS_SUBPART10,
 	BLIS_SUBPART20,
@@ -801,6 +811,80 @@ typedef enum
 #define BLIS_NUM_UKR_IMPL_TYPES 4
 
 
+#if 0
+typedef enum
+{
+	// RV = row-stored, contiguous vector-loading
+	// RG = row-stored, non-contiguous gather-loading
+	// CV = column-stored, contiguous vector-loading
+	// CG = column-stored, non-contiguous gather-loading
+
+	// RD = row-stored, dot-based
+	// CD = col-stored, dot-based
+
+	// RC = row-stored, column-times-column
+	// CR = column-stored, row-times-row
+
+	// GX = general-stored generic implementation
+
+	BLIS_GEMMSUP_RV_UKR = 0,
+	BLIS_GEMMSUP_RG_UKR,
+	BLIS_GEMMSUP_CV_UKR,
+	BLIS_GEMMSUP_CG_UKR,
+
+	BLIS_GEMMSUP_RD_UKR,
+	BLIS_GEMMSUP_CD_UKR,
+
+	BLIS_GEMMSUP_RC_UKR,
+	BLIS_GEMMSUP_CR_UKR,
+
+	BLIS_GEMMSUP_GX_UKR,
+} l3sup_t;
+
+#define BLIS_NUM_LEVEL3_SUP_UKRS 9
+#endif
+
+
+typedef enum
+{
+	// 3-operand storage combinations
+	BLIS_RRR = 0,
+	BLIS_RRC, // 1
+	BLIS_RCR, // 2
+	BLIS_RCC, // 3
+	BLIS_CRR, // 4
+	BLIS_CRC, // 5
+	BLIS_CCR, // 6
+	BLIS_CCC, // 7
+	BLIS_XXX, // 8
+
+#if 0
+	BLIS_RRG,
+	BLIS_RCG,
+	BLIS_RGR,
+	BLIS_RGC,
+	BLIS_RGG,
+	BLIS_CRG,
+	BLIS_CCG,
+	BLIS_CGR,
+	BLIS_CGC,
+	BLIS_CGG,
+	BLIS_GRR,
+	BLIS_GRC,
+	BLIS_GRG,
+	BLIS_GCR,
+	BLIS_GCC,
+	BLIS_GCG,
+	BLIS_GGR,
+	BLIS_GGC,
+	BLIS_GGG,
+#endif
+} stor3_t;
+
+#define BLIS_NUM_3OP_RC_COMBOS 9
+//#define BLIS_NUM_3OP_RCG_COMBOS 27
+
+
 #if 0
 typedef enum
 {
@@ -863,8 +947,10 @@ typedef enum
 	BLIS_MC,
 	BLIS_KC,
 	BLIS_NC,
+
 	BLIS_M2, // level-2 blocksize in m dimension
 	BLIS_N2, // level-2 blocksize in n dimension
+
 	BLIS_AF, // level-1f axpyf fusing factor
 	BLIS_DF, // level-1f dotxf fusing factor
 	BLIS_XF, // level-1f dotxaxpyf fusing factor
@@ -875,6 +961,19 @@ typedef enum
 #define BLIS_NUM_BLKSZS 11
 
 
+// -- Threshold ID type --
+
+typedef enum
+{
+	BLIS_MT = 0, // level-3 small/unpacked matrix threshold in m dimension
+	BLIS_NT,     // level-3 small/unpacked matrix threshold in n dimension
+	BLIS_KT      // level-3 small/unpacked matrix threshold in k dimension
+
+} threshid_t;
+
+#define BLIS_NUM_THRESH 3
+
+
 // -- Architecture ID type --
 
 // NOTE: This typedef enum must be kept up-to-date with the arch_t
@@ -900,12 +999,14 @@ typedef enum
 	BLIS_ARCH_BULLDOZER,
 
 	// ARM
+	BLIS_ARCH_THUNDERX2,
 	BLIS_ARCH_CORTEXA57,
 	BLIS_ARCH_CORTEXA53,
 	BLIS_ARCH_CORTEXA15,
 	BLIS_ARCH_CORTEXA9,
 
 	// IBM/Power
+	BLIS_ARCH_POWER9,
 	BLIS_ARCH_POWER7,
 	BLIS_ARCH_BGQ,
 
@@ -914,19 +1015,25 @@ typedef enum
 
 } arch_t;
 
-#define BLIS_NUM_ARCHS 18
+#define BLIS_NUM_ARCHS 20
 
 
 //
 // -- BLIS misc. structure types -----------------------------------------------
 //
 
+// These headers must be included here (or earlier) because definitions they
+// provide are needed in the pool_t and related structs.
+#include "bli_pthread.h"
+#include "bli_malloc.h"
+
 // -- Pool block type --
 
 typedef struct
 {
-	void* buf_sys;
-	void* buf_align;
+	void*     buf;
+	siz_t     block_size;
+
 } pblk_t;
 
 
@@ -934,31 +1041,57 @@ typedef struct
 
 typedef struct
 {
-	pblk_t* block_ptrs;
-	dim_t   block_ptrs_len;
+	void*     block_ptrs;
+	dim_t     block_ptrs_len;
 
-	dim_t   top_index;
-	dim_t   num_blocks;
+	dim_t     top_index;
+	dim_t     num_blocks;
+
+	siz_t     block_size;
+	siz_t     align_size;
+
+	malloc_ft malloc_fp;
+	free_ft   free_fp;
 
-	siz_t   block_size;
-	siz_t   align_size;
 } pool_t;
 
 
-// -- Memory broker object type --
+// -- Array type --
 
-// These headers must be included here (or earlier) because definitions they
-// provide are needed in the membrk_t struct.
-#include "bli_pthread.h"
-#include "bli_malloc.h"
+typedef struct
+{
+	void*     buf;
+
+	siz_t     num_elem;
+	siz_t     elem_size;
+
+} array_t;
+
+
+// -- Locked pool-of-arrays-of-pools type --
+
+typedef struct
+{
+	bli_pthread_mutex_t mutex;
+	pool_t              pool;
+
+	siz_t               def_array_len;
+
+} apool_t;
+
+
+// -- packing block allocator: Locked set of pools type --
 
 typedef struct membrk_s
 {
 	pool_t              pools[3];
 	bli_pthread_mutex_t mutex;
 
+	// These fields are used for general-purpose allocation.
+	siz_t               align_size;
 	malloc_ft           malloc_fp;
 	free_ft             free_fp;
+
 } membrk_t;
 
 
@@ -969,7 +1102,6 @@ typedef struct mem_s
 	pblk_t    pblk;
 	packbuf_t buf_type;
 	pool_t*   pool;
-	membrk_t* membrk;
 	siz_t     size;
 } mem_t;
 
@@ -981,7 +1113,8 @@ struct cntl_s
 	// Basic fields (usually required).
 	opid_t         family;
 	bszid_t        bszid;
-	void*          var_func;
+	void_fp        var_func;
+	struct cntl_s* sub_prenode;
 	struct cntl_s* sub_node;
 
 	// Optional fields (needed only by some operations such as packm).
@@ -1013,7 +1146,7 @@ typedef struct blksz_s
 typedef struct func_s
 {
 	// Kernel function address.
-	void*  ptr[BLIS_NUM_FP_TYPES];
+	void_fp ptr[BLIS_NUM_FP_TYPES];
 
 } func_t;
 
@@ -1104,6 +1237,71 @@ typedef struct obj_s
 	dim_t         n_panel;  // n dimension of a "full" panel
 } obj_t;
 
+// Pre-initializors. Things that must be set afterwards:
+// - root object pointer
+// - info bitfields: dt, target_dt, exec_dt, comp_dt
+// - info2 bitfields: scalar_dt
+// - elem_size
+// - dims, strides
+// - buffer
+// - internal scalar buffer (must always set imaginary component)
+
+#define BLIS_OBJECT_INITIALIZER \
+{ \
+	.root      = NULL, \
+\
+	.off       = { 0, 0 }, \
+	.dim       = { 0, 0 }, \
+	.diag_off  = 0, \
+\
+	.info      = 0x0 | BLIS_BITVAL_DENSE      | \
+	                   BLIS_BITVAL_GENERAL, \
+	.info2     = 0x0, \
+	.elem_size = sizeof( float ), /* this is changed later. */ \
+\
+	.buffer    = NULL, \
+	.rs        = 0, \
+	.cs        = 0, \
+	.is        = 1,  \
+\
+	.scalar    = { 0.0, 0.0 }, \
+\
+	.m_padded  = 0, \
+	.n_padded  = 0, \
+	.ps        = 0, \
+	.pd        = 0, \
+	.m_panel   = 0, \
+	.n_panel   = 0  \
+}
+
+#define BLIS_OBJECT_INITIALIZER_1X1 \
+{ \
+	.root      = NULL, \
+\
+	.off       = { 0, 0 }, \
+	.dim       = { 1, 1 }, \
+	.diag_off  = 0, \
+\
+	.info      = 0x0 | BLIS_BITVAL_DENSE      | \
+	                   BLIS_BITVAL_GENERAL, \
+	.info2     = 0x0, \
+	.elem_size = sizeof( float ), /* this is changed later. */ \
+\
+	.buffer    = NULL, \
+	.rs        = 0, \
+	.cs        = 0, \
+	.is        = 1,  \
+\
+	.scalar    = { 0.0, 0.0 }, \
+\
+	.m_padded  = 0, \
+	.n_padded  = 0, \
+	.ps        = 0, \
+	.pd        = 0, \
+	.m_panel   = 0, \
+	.n_panel   = 0  \
+}
+
 // Define these macros here since they must be updated if contents of
 // obj_t changes.
 
@@ -1170,6 +1368,39 @@ static void bli_obj_init_subpart_from( obj_t* a, obj_t* b )
 	b->n_panel   = a->n_panel;
 }
 
+// Initializors for global scalar constants.
+// NOTE: These must remain cpp macros since they are initializor
+// expressions, not functions.
+
+#define bli_obj_init_const( buffer0 ) \
+{ \
+	.root      = NULL, \
+\
+	.off       = { 0, 0 }, \
+	.dim       = { 1, 1 }, \
+	.diag_off  = 0, \
+\
+	.info      = 0x0 | BLIS_BITVAL_CONST_TYPE | \
+	                   BLIS_BITVAL_DENSE      | \
+	                   BLIS_BITVAL_GENERAL, \
+	.info2     = 0x0, \
+	.elem_size = sizeof( constdata_t ), \
+\
+	.buffer    = buffer0, \
+	.rs        = 1, \
+	.cs        = 1, \
+	.is        = 1  \
+}
+
+#define bli_obj_init_constdata( val ) \
+{ \
+	.s =           ( float  )val, \
+	.d =           ( double )val, \
+	.c = { .real = ( float  )val, .imag = 0.0f }, \
+	.z = { .real = ( double )val, .imag = 0.0 }, \
+	.i =           ( gint_t )val, \
+}
+
 
 // -- Context type --
 
@@ -1182,6 +1413,12 @@ typedef struct cntx_s
 	func_t    l3_nat_ukrs[ BLIS_NUM_LEVEL3_UKRS ];
 	mbool_t   l3_nat_ukrs_prefs[ BLIS_NUM_LEVEL3_UKRS ];
 
+	blksz_t   l3_sup_thresh[ BLIS_NUM_THRESH ];
+	void*     l3_sup_handlers[ BLIS_NUM_LEVEL3_OPS ];
+	blksz_t   l3_sup_blkszs[ BLIS_NUM_BLKSZS ];
+	func_t    l3_sup_kers[ BLIS_NUM_3OP_RC_COMBOS ];
+	mbool_t   l3_sup_kers_prefs[ BLIS_NUM_3OP_RC_COMBOS ];
+
 	func_t    l1f_kers[ BLIS_NUM_LEVEL1F_KERS ];
 	func_t    l1v_kers[ BLIS_NUM_LEVEL1V_KERS ];
 
@@ -1193,7 +1430,6 @@ typedef struct cntx_s
 	pack_t    schema_b_panel;
 	pack_t    schema_c_panel;
 
-	membrk_t* membrk;
 } cntx_t;
 
 
@@ -1201,9 +1437,21 @@ typedef struct cntx_s
 
 typedef struct rntm_s
 {
+	// "External" fields: these may be queried by the end-user.
 	dim_t     num_threads;
 	dim_t     thrloop[ BLIS_NUM_LOOPS ];
 
+	// "Internal" fields: these should not be exposed to the end-user.
+
+	// The small block pool, which is attached in the l3 thread decorator.
+	pool_t*   sba_pool;
+
+	// The packing block allocator, which is attached in the l3 thread decorator.
+	membrk_t* membrk;
+
+	// A switch to enable/disable small/unpacked matrix handling in level-3 ops.
+	bool_t    l3_sup;
+
 } rntm_t;
 
 
@@ -1290,28 +1538,31 @@ typedef enum
 	// Buffer-specific errors 
 	BLIS_EXPECTED_NONNULL_OBJECT_BUFFER        = (-110),
 
-	// Memory allocator errors
-	BLIS_INVALID_PACKBUF                       = (-120),
-	BLIS_EXHAUSTED_CONTIG_MEMORY_POOL          = (-122),
-	BLIS_INSUFFICIENT_STACK_BUF_SIZE           = (-123),
-	BLIS_ALIGNMENT_NOT_POWER_OF_TWO            = (-124),
-	BLIS_ALIGNMENT_NOT_MULT_OF_PTR_SIZE        = (-125),
+	// Memory errors
+	BLIS_MALLOC_RETURNED_NULL                  = (-120),
+
+	// Internal memory pool errors
+	BLIS_INVALID_PACKBUF                       = (-130),
+	BLIS_EXHAUSTED_CONTIG_MEMORY_POOL          = (-131),
+	BLIS_INSUFFICIENT_STACK_BUF_SIZE           = (-132),
+	BLIS_ALIGNMENT_NOT_POWER_OF_TWO            = (-133),
+	BLIS_ALIGNMENT_NOT_MULT_OF_PTR_SIZE        = (-134),
 
 	// Object-related errors
-	BLIS_EXPECTED_OBJECT_ALIAS                 = (-130),
+	BLIS_EXPECTED_OBJECT_ALIAS                 = (-140),
 
 	// Architecture-related errors
-	BLIS_INVALID_ARCH_ID                       = (-140),
+	BLIS_INVALID_ARCH_ID                       = (-150),
 
 	// Blocksize-related errors
-	BLIS_MC_DEF_NONMULTIPLE_OF_MR              = (-150),
-	BLIS_MC_MAX_NONMULTIPLE_OF_MR              = (-151),
-	BLIS_NC_DEF_NONMULTIPLE_OF_NR              = (-152),
-	BLIS_NC_MAX_NONMULTIPLE_OF_NR              = (-153),
-	BLIS_KC_DEF_NONMULTIPLE_OF_KR              = (-154),
-	BLIS_KC_MAX_NONMULTIPLE_OF_KR              = (-155),
+	BLIS_MC_DEF_NONMULTIPLE_OF_MR              = (-160),
+	BLIS_MC_MAX_NONMULTIPLE_OF_MR              = (-161),
+	BLIS_NC_DEF_NONMULTIPLE_OF_NR              = (-162),
+	BLIS_NC_MAX_NONMULTIPLE_OF_NR              = (-163),
+	BLIS_KC_DEF_NONMULTIPLE_OF_KR              = (-164),
+	BLIS_KC_MAX_NONMULTIPLE_OF_KR              = (-165),
 
-	BLIS_ERROR_CODE_MAX                        = (-160)
+	BLIS_ERROR_CODE_MAX                        = (-170)
 } err_t;
 
 #endif
